#!/usr/bin/env python3

from .db import Deck, Card, get_card_by_id, get_deck_by_id

from .port import export_deck, import_deck

from .spacedr import create_card, edit_card, delete_card, create_deck
from .spacedr import edit_deck, delete_deck

from .schedule import update_card, get_cards_to_study, get_cards_to_review

from . import errors

from .__version__ import __title__, __description__, __url__, __version__
from .__version__ import __author__, __license__, __copyright__

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s:'
                                  '%(funcName)s > %(message)s',
                              datefmt='%d/%m %I:%M:%S')

handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
handler.setFormatter(formatter)

logger.addHandler(handler)



#!/usr/bin/env python3

import os
import pytest
import spacedr


@pytest.fixture(scope='session', autouse=True)
def delete_db_after_test(request):

    spacedr.db.db_init()

    def teardown() -> None:
        db_path = spacedr.db.get_db_path()
        os.remove(db_path)

    request.addfinalizer(teardown)



#!/usr/bin/env python3

import logging
from typing import List, Any
from datetime import datetime, timedelta

from .db import Card, Deck, get_session

logger = logging.getLogger(__name__)

THRESHOLDS = [
    timedelta(seconds=40),
    timedelta(hours=3),
    timedelta(hours=8),
    timedelta(days=1),
    timedelta(days=3),
    timedelta(days=7),
    timedelta(days=30)
]


@get_session(use_obj_session=True, commit=False, close=False)
def update_card(card: Card, answer: Any) -> None:
    '''Update card data.

    Usage::

        >>> spacedr.update_card(card, answer='43')
    '''
    card.reviews_count += 1

    if answer in card.answers:
        card.correct_count += 1
        level = card.level + 1
        card.last_correct = True
    else:
        level = card.level - 1
        if level < 0:  # If first wrong
            level = 0
        card.last_correct = False

    card.level = level

    if level == 7 or (card.reviews_count == 1 and card.level == 1):
        card.practiced = True
        card.due_date = None

    if answer not in card.answers:  # If wrong, review now
        level = 0

    card.due_date = datetime.now() + THRESHOLDS[level]

    session.commit()
    logger.info(f'Updated card with id: {card.id} to level: {card.level}, '
                f'due_date: {card.due_date}')


@get_session()
def get_cards_to_study(deck: Deck, num: int) -> List[Card]:
    '''Get cards that haven't been studied, limit quantity by ``num``.'''
    return list(session.query(Card)
                .filter(Card.deck_id == deck.id)
                .filter(Card.reviews_count == 0)
                .order_by(Card.id.asc())
                .limit(num))


@get_session()
def get_cards_to_review(deck: Deck) -> List[Card]:
    '''Get cards that can be reviewed.'''
    return list(session.query(Card)
                .filter(Card.deck_id == deck.id)
                .filter(Card.practiced == False)
                .filter(Card.reviews_count >= 1)
                .filter(Card.due_date <= datetime.now())
                .order_by(Card.due_date.asc()))



#!/usr/bin/env python3

__title__ = 'spacedr'
__description__ = 'Spaced Repetition API'
__url__ = 'http://github.com/sttlr/spacedr'
__version__ = '1.0.1'
__author__ = 'sttlr'
__license__ = 'MIT'
__copyright__ = 'Copyright 2018, sttlr'



#!/usr/bin/env python3

import os
import logging
from typing import List, Callable, Optional, Any
from functools import wraps
from datetime import datetime

from sqlalchemy import create_engine, inspect
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, ForeignKey, Integer, String, DateTime, Boolean
from sqlalchemy import PickleType, Float
from sqlalchemy.ext.mutable import MutableList, MutableDict

from .errors import DeckNotFoundError, CardNotFoundError

logger = logging.getLogger(__name__)


def get_db_path() -> str:
    '''Get a path to the database.'''
    return os.path.join(os.path.join(os.getcwd(), 'spacedr'), 'spacedr.sqlite')


engine = create_engine('sqlite:///%s' % os.path.abspath(get_db_path()))
session_maker = sessionmaker(bind=engine, expire_on_commit=False)
Base = declarative_base()


def get_session(commit: bool=True, close: bool=True,
                use_obj_session: bool=False) -> Callable:
    '''Get a sqlalchemy ``session`` object applying parameters via decorator.

       :param commit: commit after making a change.
       :param close: finally close the session.
       :param use_obj_session: use object's session to make a change.
    '''

    def decorator(func) -> Callable:

        @wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            if use_obj_session:
                session = inspect(args[0]).session
            else:
                session = session_maker()

            try:
                func.__globals__['session'] = session
                result = func(*args, **kwargs)
                if commit:
                    session.commit()
                return result
            except Exception as e:
                session.rollback()
                logger.exception(e)
                raise
            finally:
                if close:
                    session.close()

        return wrapper
    return decorator


class Card(Base):
    '''A class used to create a card.'''

    __tablename__ = 'card'

    id: int = Column('id', Integer, primary_key=True, autoincrement=True)
    deck_id: int = Column('deck_id', Integer, ForeignKey('deck.id'),
                          nullable=False)
    question: str = Column('question', String, nullable=False)
    examples: List[str] = Column('examples', MutableList.as_mutable(PickleType),
                                 nullable=True)
    answers: List[str] = Column('answers', MutableList.as_mutable(PickleType),
                                nullable=False)
    level: int = Column('level', Integer, nullable=False, default=0)
    due_date: datetime = Column('due_date', DateTime, nullable=True,
                                default=None)
    reviews_count: int = Column('reviews_count', Integer, nullable=False,
                                default=0)
    correct_count: int = Column('correct_count', Integer, nullable=False,
                                default=0)
    last_correct: bool = Column('last_correct', Boolean, nullable=True,
                                default=None)
    practiced: bool = Column('practiced', Boolean, nullable=False,
                             default=False)
    other: dict = Column('other', MutableDict.as_mutable(PickleType),
                         nullable=True, default=None)


class Deck(Base):
    '''A class used to create a deck.'''

    __tablename__ = 'deck'

    id: int = Column('id', Integer, primary_key=True)
    name: str = Column('name', String, nullable=False)
    cards: List[Card] = relationship(Card, cascade='all', backref='deck')
    description: Optional[str] = Column('description', String)


def db_init() -> None:
    '''Initialise the database.'''
    os.makedirs(os.path.dirname(get_db_path()), exist_ok=True)
    Base.metadata.create_all(engine)
    logger.debug('The database was successfully initialised')


@get_session(close=False)
def get_deck_by_id(id: int) -> Deck:
    '''Get deck by id.

    :raises errors.DeckNotFoundError: No deck with given id.
    '''
    deck = session.query(Deck).filter(Deck.id == id).one_or_none()
    if not deck:
        raise DeckNotFoundError(f'There is no deck with id: {id}')
    logger.debug(f'Found a deck with id: {deck.id}')
    return deck


@get_session(close=False)
def get_card_by_id(id: int) -> Card:
    '''Get card by id.

    :raises errors.CardNotFoundError: No card with given id.
    '''
    card = session.query(Card).filter(Card.id == id).one_or_none()
    if not card:
        raise CardNotFoundError(f'There is no card with id: {id}')
    logger.debug(f'Found a card with id: {card.id}')
    return card



#!/usr/bin/env python3

from typing import List
import pytest
import spacedr

'''Run tests sequentially. Other way, it fails.'''


@pytest.mark.parametrize('name,description,id', [('test', None, 1),
                                                 ('what', 'nothing', 2)],)
def test_create_deck(name: str, description: str, id: int) -> None:
    spacedr.create_deck(name, description)
    deck = spacedr.get_deck_by_id(id)

    assert deck.name == name
    assert deck.description == description


@pytest.mark.parametrize('name,description,id', [('what', 'nothing', 1),
                                                 ('test', None, 2)])
def test_edit_deck(name: str, description: str, id: int) -> None:
    deck = spacedr.get_deck_by_id(id)
    spacedr.edit_deck(deck, name=name, description=description)

    assert deck.name == name
    assert deck.description == description or 'nothing'


@pytest.mark.parametrize('id', [2])
def test_delete_deck(id: int) -> None:
    deck = spacedr.get_deck_by_id(id)
    spacedr.delete_deck(deck)

    with pytest.raises(spacedr.errors.DeckNotFoundError):
        spacedr.get_deck_by_id(id)


@pytest.mark.parametrize('deck_id,card_id,question,answers,examples,other',
                         [(1, 1, 'meaning of life', [42, '42'], ['42dot'],
                           {'key1': 'val1', 'key2': 'val2'}),
                          (1, 2, 'nothing else than 0', [0], ['0', 0.0, 0],
                           {'key3': ['val3', 'val']}),
                          (1, 3, 'just third', [3], ['only 3', 3], {'4': 5}),
                          (1, 4, 'only thing you need to know', [8], ['what'],
                           {'hello': ['gbye']}),
                          (1, 5, 'five.', [5], [5, 5.5], {'key': {'k': 'v'}})])
def test_create_card(deck_id: int, card_id: int, question: str,
                     answers: List[str], examples: List[str],
                     other: dict) -> None:
    deck = spacedr.get_deck_by_id(deck_id)
    spacedr.create_card(deck, question, answers, examples, other)

    card = spacedr.get_card_by_id(card_id)

    assert card.question == question
    assert card.answers == answers
    assert card.examples == examples
    assert card.other == other


@pytest.mark.parametrize('deck_id,card_id,question,answers,examples,other',
                         [(1, 1, 'nothing else than 0', [0], ['4'], {'n': 5}),
                          (1, 2, 'life meaning', [42, '42'], [2], {'l': 'f'})])
def test_edit_card(deck_id: int, card_id: int, question: str,
                   answers: List[str], examples: List[str],
                   other: dict) -> None:
    card = spacedr.get_card_by_id(card_id)
    spacedr.edit_card(card, deck_id=deck_id, question=question,
                      answers=answers, examples=examples, other=other)

    assert card.deck_id == deck_id
    assert card.question == question
    assert card.answers == answers
    assert card.examples == examples
    assert card.other == other


@pytest.mark.parametrize('id', [3])
def test_delete_card(id: int) -> None:
    card = spacedr.get_card_by_id(id)
    spacedr.delete_card(card)

    with pytest.raises(spacedr.errors.CardNotFoundError):
        spacedr.get_card_by_id(id)


@pytest.mark.parametrize('id,answer', [(1, '11'), (2, 42), (2, '42')])
def test_update_card(id: int, answer: str) -> None:
    card = spacedr.get_card_by_id(id)
    old_data = {'due_date': card.due_date, 'reviews_count': card.reviews_count}

    spacedr.update_card(card, answer)

    assert card.due_date != old_data['due_date']
    assert card.reviews_count == old_data['reviews_count'] + 1


@pytest.mark.parametrize('deck_id,num', [(1, 1), (1, 3)])
def test_get_cards_to_study(deck_id: int, num: int) -> None:
    deck = spacedr.get_deck_by_id(deck_id)
    cards_to_study = spacedr.get_cards_to_study(deck, num)

    assert len(cards_to_study) <= num

    for card in cards_to_study:
        assert card.deck_id == deck.id
        assert card.level == 0
        assert card.due_date == None
        assert card.reviews_count == 0


@pytest.mark.parametrize('deck_id', [1])
def test_get_cards_to_review(deck_id: int) -> None:
    deck = spacedr.get_deck_by_id(deck_id)
    cards_to_review = spacedr.get_cards_to_review(deck)

    for card in cards_to_review:
        assert card.deck_id == deck.id
        assert card.due_date != None
        assert card.reviews_count != 0


@pytest.mark.parametrize('id', [10, 42])
def test_error_deck_not_found(id: int) -> None:
    with pytest.raises(spacedr.errors.DeckNotFoundError):
        spacedr.get_deck_by_id(id)


@pytest.mark.parametrize('id', [15, 66])
def test_error_card_not_found(id: int) -> None:
    with pytest.raises(spacedr.errors.CardNotFoundError):
        spacedr.get_card_by_id(id)


@pytest.mark.parametrize('card_id', [(2)])
def test_error_card_already_exists(card_id: int) -> None:
    card = spacedr.get_card_by_id(card_id)
    deck = spacedr.get_deck_by_id(card.deck_id)

    with pytest.raises(spacedr.errors.CardAlreadyExists):
        try:
            spacedr.create_card(deck, question=card.question,
                                answers=card.answers)
        except spacedr.errors.CardAlreadyExists:
            raise



#!/usr/bin/env python3

import io
import os
import sys
from shutil import rmtree

from setuptools import setup, find_packages, Command
from setuptools.command.test import test as TestCommand

# Package meta-data.
REQUIRES_PYTHON = '>=3.5'

# What packages are required for this module to be executed?
REQUIRED = ['sqlalchemy==1.2.12'
            ]
TESTS_REQUIRE = ['pytest==3.9.3']

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, 'spacedr', '__version__.py')) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('py setup.py sdist')

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


# Where the magic happens:
setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    author=about['__author__'],
    python_requires=REQUIRES_PYTHON,
    url=about['__url__'],
    packages=find_packages(exclude=('tests',)),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license=about['__license__'],
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    tests_require=TESTS_REQUIRE,
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
        'test': PyTest
    },
)



#!/usr/bin/env python3

import logging
from typing import List

from .db import Card, Deck, get_session, get_deck_by_id
from .errors import CardAlreadyExists, DeckNotFoundError

logger = logging.getLogger(__name__)


@get_session(use_obj_session=True, commit=False, close=False)
def create_card(deck: Deck, question: str, answers: List[str],
                examples: List[str]=None, other: dict=None) -> None:
    '''Create a card.

    :raises errors.CardAlreadyExists: There is a card in the same deck with the same name.

    Usage::

        >>> spacedr.create_card(deck,
        ...                     question='What is the meaning of life?',
        ...                     answers=[42, '42'])
    '''
    if session.query(Card).filter(Card.deck_id == deck.id)\
       .filter(Card.question == question).one_or_none():
        raise CardAlreadyExists(f'There is a card in the same deck with the '
                                f'same question: {question}')

    card = Card(deck_id=deck.id, question=question, answers=answers,
                examples=examples, other=other)
    deck.cards.append(card)

    session.add(card)
    session.commit()
    logger.info(f'Created card with id: {card.id}')


@get_session(use_obj_session=True, commit=False, close=False)
def edit_card(card: Card, **kwargs) -> None:
    '''Edit card's deck_id, question and answers.

    :param int deck_id: move the card to a different deck.question.
    :param str question: change the card's question.
    :param list answers: set the new answers to a card.

    :raises errors.DeckNotFoundError: can't find deck with given ``deck_id``

    Usage:

        >>> spacedr.edit_card(card, deck_id=deck_id, question='What is life?',
        ...                   answers=[42])
    '''
    deck_id = kwargs.get('deck_id', None) or card.deck_id
    try:
        get_deck_by_id(deck_id)
    except DeckNotFoundError:
        raise

    question = kwargs.get('question', None) or card.question
    answers = kwargs.get('answers', None) or card.answers
    examples = kwargs.get('examples', None) or card.examples
    other = kwargs.get('other', None) or card.other

    card.deck_id = deck_id
    card.question = question
    card.answers = answers
    card.examples = examples
    card.other = other

    session.commit()
    logger.info(f'Edited card with id: {card.id}. Assigned it to the deck '
                f'with id: {deck_id}, question: {card.question}, answers: '
                f'{card.answers}, examples: {card.examples}, other: '
                f'{card.other}')


@get_session(use_obj_session=True)
def delete_card(card: Card) -> None:
    '''Delete a card.'''
    session.delete(card)
    logger.info(f'Deleted card with id: {card.id}')


@get_session(commit=False)
def create_deck(name: str, description: str=None) -> None:
    '''Create a deck.

    Usage::

        >>> spacedr.create_deck(name='Test Deck', description='Just a test.')
    '''
    deck = Deck(name=name, description=description)
    session.add(deck)
    session.commit()
    logger.info(f'Created deck with id: {deck.id}')


@get_session(use_obj_session=True, commit=False, close=False)
def edit_deck(deck: Deck, **kwargs) -> None:
    '''Edit deck's name, description.

    :param str name: new name to the deck.
    :param str description: assign description to the deck.

    Usage::

        >>> spacedr.edit_deck(deck, name='test2', description='new one')
    '''
    name = kwargs.get('name', None) or deck.name
    description = kwargs.get('description', None) or deck.description

    deck.name = name
    deck.description = description

    session.commit()
    logger.info(f'Edited deck with id: {deck.id}. name: {deck.name}, '
                f'description: {deck.description}')


@get_session(use_obj_session=True)
def delete_deck(deck: Deck) -> None:
    '''Delete a deck.'''
    session.delete(deck)
    logger.info(f'Deleted deck with id: {deck.id}')



#!/usr/bin/env python3

import logging
import json
from typing import IO, Any

from .db import Deck, Card, get_session

logger = logging.getLogger(__name__)


def export_deck(deck: Deck, fd: IO[Any]) -> None:
    '''Export a deck to a file descriptor.

    Usage::

        >>> with open('mydeck.json', 'w+') as fd:
        ...     spacedr.export_deck(deck, fd)
    '''
    json.dump({'name': deck.name,
               'description': deck.description,
               'cards': [{'question': card.question,
                          'answers': card.answers,
                          } for card in deck.cards]
               }, fd, separators=(',', ':'), check_circular=False)
    logger.info('The deck was successfully exported.')


@get_session()
def import_deck(fd: IO[Any]) -> None:
    '''Import a deck from a file descriptor as the new one.
    The old won't be removed.

    Usage::

        >>> with open('mydeck.json') as fd:
        ...     spacedr.import_deck(fd)
    '''
    try:
        deck_obj = json.load(fd)
    except Exception as e:
        raise e

    deck = Deck(name=deck_obj['name'], description=deck_obj['description'])
    session.commit()

    for card_obj in deck_obj['cards']:
        card = Card(deck_id=deck.id, question=card_obj['question'],
                    answers=card_obj['answers'])
        deck.cards.append(card)

    session.add(deck)
    logger.info('The deck was successfully imported.')



class DeckNotFoundError(Exception):
    '''No deck with given id.'''


class CardNotFoundError(Exception):
    '''No card with given id.'''


class CardAlreadyExists(Exception):
    '''There is a card with the same question.'''
